@@ -101,6 +101,8 @@ static const char *ui_description =
         "      <menuitem action='Home'/>"
         "      <menuitem action='Back'/>"
         "      <menuitem action='Forward'/>"
+        "      <menuitem action='Prev'/>"
+        "      <menuitem action='Next'/>"
         "      <separator/>"
         "      <menuitem action='ZoomIn'/>"
         "      <menuitem action='ZoomReset'/>"
@@ -118,6 +120,8 @@ static const char *ui_description =
         "    <toolitem action='Back'/>"
         "    <toolitem action='Forward'/>"
         "    <toolitem action='Home'/>"
+        "    <toolitem action='Prev'/>"
+        "    <toolitem action='Next'/>"
         "    <toolitem action='ZoomIn'/>"
         "    <toolitem action='ZoomReset'/>"
         "    <toolitem action='ZoomOut'/>"
@@ -136,7 +140,6 @@ static void chmsee_finalize(GObject *);
 static void chmsee_dispose(GObject *);
 
 static gboolean delete_cb(GtkWidget *, GdkEvent *, Chmsee *);
-static void destroy_cb(GtkWidget *, Chmsee *);
 static gboolean window_state_event_cb(Chmsee *, GdkEventWindowState *);
 static gboolean configure_event_cb(GtkWidget *, GdkEventConfigure *, Chmsee *);
 static void book_model_changed_cb(Chmsee *, CsChmfile *, const gchar *);
@@ -146,29 +149,31 @@ static void book_message_notify_cb(Chmsee *, GParamSpec *, CsBook *);
 static void open_file_response_cb(GtkWidget *, gint, Chmsee *);
 static void about_response_cb(GtkDialog *, gint, gpointer);
 
-static void on_open_file(GtkWidget *, Chmsee *);
+static void on_open_file(GtkAction *, Chmsee *);
 static void on_recent_files(GtkRecentChooser *, Chmsee *);
-static void on_open_new_tab(GtkWidget *, Chmsee *);
-static void on_close_current_tab(GtkWidget *, Chmsee *);
+static void on_open_new_tab(GtkAction *, Chmsee *);
+static void on_close_current_tab(GtkAction *, Chmsee *);
 
-static void on_menu_file(GtkWidget *, Chmsee *);
-static void on_menu_edit(GtkWidget *, Chmsee *);
-static void on_home(GtkWidget *, Chmsee *);
-static void on_back(GtkWidget *, Chmsee *);
-static void on_forward(GtkWidget *, Chmsee *);
-static void on_zoom_in(GtkWidget *, Chmsee *);
-static void on_zoom_reset(GtkWidget *, Chmsee *);
-static void on_zoom_out(GtkWidget *, Chmsee *);
-static void on_setup(GtkWidget *, Chmsee *);
-static void on_about(GtkWidget *);
-static void on_copy(GtkWidget *, Chmsee *);
-static void on_select_all(GtkWidget *, Chmsee *);
-static void on_find(GtkWidget *, Chmsee *);
-static void on_keyboard_escape(GtkWidget *, Chmsee *);
-static void on_fullscreen_toggled(GtkWidget *, Chmsee *);
-static void on_sidepane_toggled(GtkWidget *, Chmsee *);
+static void on_quit(GtkAction *, Chmsee *);
+static void on_menu_file(GtkAction *, Chmsee *);
+static void on_menu_edit(GtkAction *, Chmsee *);
+static void on_home(GtkAction *, Chmsee *);
+static void on_back(GtkAction *, Chmsee *);
+static void on_forward(GtkAction *, Chmsee *);
+static void on_prev(GtkAction *, Chmsee *);
+static void on_next(GtkAction *, Chmsee *);
+static void on_zoom_in(GtkAction *, Chmsee *);
+static void on_zoom_reset(GtkAction *, Chmsee *);
+static void on_zoom_out(GtkAction *, Chmsee *);
+static void on_setup(GtkAction *, Chmsee *);
+static void on_about(GtkAction *);
+static void on_copy(GtkAction *, Chmsee *);
+static void on_select_all(GtkAction *, Chmsee *);
+static void on_find(GtkAction *, Chmsee *);
+static void on_keyboard_escape(GtkAction *, Chmsee *);
+static void on_fullscreen_toggled(GtkToggleAction *, Chmsee *);
+static void on_sidepane_toggled(GtkToggleAction *, Chmsee *);
 
-static void chmsee_quit(Chmsee *);
 static void populate_windows(Chmsee *);
 static void chmsee_set_fullscreen(Chmsee *, gboolean);
 static void show_sidepane(Chmsee *);
@@ -193,7 +198,7 @@ static const GtkActionEntry entries[] = {
 
         { "NewTab", NULL, N_("New _Tab"), "<control>T", NULL, G_CALLBACK(on_open_new_tab)},
         { "CloseTab", NULL, N_("_Close Tab"), "<control>W", NULL, G_CALLBACK(on_close_current_tab)},
-        { "Exit", GTK_STOCK_QUIT, N_("E_xit"), "<control>Q", N_("Exit ChmSee"), G_CALLBACK(destroy_cb)},
+        { "Exit", GTK_STOCK_QUIT, N_("E_xit"), "<control>Q", N_("Exit ChmSee"), G_CALLBACK(on_quit)},
 
         { "Copy", NULL, N_("_Copy"), NULL, NULL, G_CALLBACK(on_copy)},
         { "SelectAll", NULL, N_("Select _All"), NULL, NULL, G_CALLBACK(on_select_all)},
@@ -205,6 +210,8 @@ static const GtkActionEntry entries[] = {
         { "Home", GTK_STOCK_HOME, N_("_Home"), NULL, NULL, G_CALLBACK(on_home)},
         { "Back", GTK_STOCK_GO_BACK, N_("_Back"), "<alt>Left", NULL, G_CALLBACK(on_back)},
         { "Forward", GTK_STOCK_GO_FORWARD, N_("_Forward"), "<alt>Right", NULL, G_CALLBACK(on_forward)},
+        { "Prev", GTK_STOCK_GO_UP, N_("_Prev"), "<alt>Up", NULL, G_CALLBACK(on_prev)},
+        { "Next", GTK_STOCK_GO_DOWN, N_("_Next"), "<alt>Down", NULL, G_CALLBACK(on_next)},
 
         { "About", GTK_STOCK_ABOUT, N_("_About"), NULL, N_("About ChmSee"), G_CALLBACK(on_about)},
 
@@ -271,10 +278,6 @@ chmsee_init(Chmsee *self)
                          "delete-event",
                          G_CALLBACK (delete_cb),
                          self);
-        g_signal_connect(G_OBJECT (self),
-                         "destroy-event",
-                         G_CALLBACK (destroy_cb),
-                         self);
 
         /* startup html render engine */
         if(!cs_html_gecko_init_system()) {
@@ -293,6 +296,20 @@ chmsee_dispose(GObject *gobject)
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
 
         if (priv->chmfile) {
+                /* save last opened page */
+                const gchar *bookfolder = cs_chmfile_get_bookfolder(priv->chmfile);
+                gchar *location = cs_book_get_location(CS_BOOK (priv->book));
+                gchar *page = g_strrstr(location, bookfolder);
+
+                if (page) {
+                        page = page + strlen(bookfolder);
+                        gchar *last_file = g_strdup_printf("%s::%s", priv->config->last_file, page);
+                        g_free(priv->config->last_file);
+                        priv->config->last_file = last_file;
+                }
+
+                g_free(location);
+
                 g_object_unref(priv->chmfile);
                 priv->chmfile = NULL;
         }
@@ -304,7 +321,7 @@ chmsee_dispose(GObject *gobject)
                 priv->ui_manager = NULL;
         }
 
-        G_OBJECT_CLASS(chmsee_parent_class)->dispose(gobject);
+        G_OBJECT_CLASS (chmsee_parent_class)->dispose(gobject);
 }
 
 static void
@@ -321,15 +338,8 @@ static gboolean
 delete_cb(GtkWidget *widget, GdkEvent *event, Chmsee *self)
 {
         g_debug("Chmsee >>> window delete");
-        chmsee_quit(self);
-        return FALSE;
-}
-
-static void
-destroy_cb(GtkWidget *widget, Chmsee *self)
-{
-        g_debug("Chmsee >>> window destroy");
-        chmsee_quit(self);
+        on_quit(NULL, self);
+        return TRUE;
 }
 
 static gboolean
@@ -403,6 +413,8 @@ book_model_changed_cb(Chmsee *self, CsChmfile *chmfile, const gchar *filename)
         gtk_action_set_sensitive(gtk_action_group_get_action(priv->action_group, "ZoomReset"), has_model);
         gtk_action_set_sensitive(gtk_action_group_get_action(priv->action_group, "Back"), has_model);
         gtk_action_set_sensitive(gtk_action_group_get_action(priv->action_group, "Forward"), has_model);
+        gtk_action_set_sensitive(gtk_action_group_get_action(priv->action_group, "Prev"), has_model);
+        gtk_action_set_sensitive(gtk_action_group_get_action(priv->action_group, "Next"), has_model);
 
         gtk_widget_set_sensitive(priv->book, has_model);
 
@@ -459,7 +471,7 @@ open_file_response_cb(GtkWidget *widget, gint response_id, Chmsee *self)
 /* Toolbar button events */
 
 static void
-on_open_file(GtkWidget *widget, Chmsee *self)
+on_open_file(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
 
@@ -515,41 +527,55 @@ on_recent_files(GtkRecentChooser *chooser, Chmsee *self)
 }
 
 static void
-on_copy(GtkWidget *widget, Chmsee *self)
+on_copy(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_copy(CS_BOOK (priv->book));
 }
 
 static void
-on_select_all(GtkWidget *widget, Chmsee *self)
+on_select_all(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_select_all(CS_BOOK (priv->book));
 }
 
 static void
-on_setup(GtkWidget *widget, Chmsee *self)
+on_setup(GtkAction *action, Chmsee *self)
 {
         setup_window_new(self);
 }
 
 static void
-on_back(GtkWidget *widget, Chmsee *self)
+on_back(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_go_back(CS_BOOK (priv->book));
 }
 
 static void
-on_forward(GtkWidget *widget, Chmsee *self)
+on_forward(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_go_forward(CS_BOOK (priv->book));
 }
 
 static void
-on_menu_file(GtkWidget *widget, Chmsee *self)
+on_prev(GtkAction *action, Chmsee *self)
+{
+        ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
+        cs_book_go_prev(CS_BOOK (priv->book));
+}
+
+static void
+on_next(GtkAction *action, Chmsee *self)
+{
+        ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
+        cs_book_go_next(CS_BOOK (priv->book));
+}
+
+static void
+on_menu_file(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         gboolean can_close_tab = cs_book_can_close_tab(CS_BOOK (priv->book));
@@ -557,7 +583,7 @@ on_menu_file(GtkWidget *widget, Chmsee *self)
 }
 
 static void
-on_menu_edit(GtkWidget *widget, Chmsee *self)
+on_menu_edit(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         gboolean can_copy = cs_book_can_copy(CS_BOOK (priv->book));
@@ -565,26 +591,26 @@ on_menu_edit(GtkWidget *widget, Chmsee *self)
 }
 
 static void
-on_home(GtkWidget *widget, Chmsee *self)
+on_home(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_homepage(CS_BOOK (priv->book));
 }
 
 static void
-on_zoom_in(GtkWidget *widget, Chmsee *self)
+on_zoom_in(GtkAction *action, Chmsee *self)
 {
         cs_book_zoom_in(CS_BOOK (CHMSEE_GET_PRIVATE (self)->book));
 }
 
 static void
-on_zoom_out(GtkWidget *widget, Chmsee *self)
+on_zoom_out(GtkAction *action, Chmsee *self)
 {
         cs_book_zoom_out(CS_BOOK (CHMSEE_GET_PRIVATE (self)->book));
 }
 
 static void
-on_zoom_reset(GtkWidget *widget, Chmsee *self)
+on_zoom_reset(GtkAction *action, Chmsee *self)
 {
         cs_book_zoom_reset(CS_BOOK (CHMSEE_GET_PRIVATE (self)->book));
 }
@@ -597,7 +623,7 @@ about_response_cb(GtkDialog *dialog, gint response_id, gpointer user_data)
 }
 
 static void
-on_about(GtkWidget *widget)
+on_about(GtkAction *action)
 {
 
         GtkBuilder *builder = gtk_builder_new();
@@ -616,7 +642,7 @@ on_about(GtkWidget *widget)
 }
 
 static void
-on_open_new_tab(GtkWidget *widget, Chmsee *self)
+on_open_new_tab(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
 
@@ -626,13 +652,13 @@ on_open_new_tab(GtkWidget *widget, Chmsee *self)
 }
 
 static void
-on_close_current_tab(GtkWidget *widget, Chmsee *self)
+on_close_current_tab(GtkAction *action, Chmsee *self)
 {
         cs_book_close_current_tab(CS_BOOK (CHMSEE_GET_PRIVATE (self)->book));
 }
 
 static void
-on_keyboard_escape(GtkWidget *widget, Chmsee *self)
+on_keyboard_escape(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
 
@@ -640,27 +666,20 @@ on_keyboard_escape(GtkWidget *widget, Chmsee *self)
 }
 
 static void
-on_fullscreen_toggled(GtkWidget *menu, Chmsee *self)
+on_fullscreen_toggled(GtkToggleAction *action, Chmsee *self)
 {
         g_return_if_fail(IS_CHMSEE(self));
 
-        gboolean active;
-        g_object_get(G_OBJECT (menu),
-                     "active", &active,
-                     NULL);
-        g_debug("Chmsee >>> enter on_fullscreen_toggled with menu.active = %d", active);
+        gboolean active = gtk_toggle_action_get_active(action);
         chmsee_set_fullscreen(self, active);
 }
 
 static void
-on_sidepane_toggled(GtkWidget *menu, Chmsee *self)
+on_sidepane_toggled(GtkToggleAction *action, Chmsee *self)
 {
         g_return_if_fail(IS_CHMSEE (self));
 
-        gboolean active;
-        g_object_get(G_OBJECT (menu),
-                     "active", &active,
-                     NULL);
+        gboolean active = gtk_toggle_action_get_active(action);
         if (active) {
                 show_sidepane(self);
         } else {
@@ -669,7 +688,7 @@ on_sidepane_toggled(GtkWidget *menu, Chmsee *self)
 }
 
 static void
-on_find(GtkWidget *widget, Chmsee *self)
+on_find(GtkAction *action, Chmsee *self)
 {
         ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
         cs_book_findbar_show(CS_BOOK (priv->book));
@@ -678,7 +697,7 @@ on_find(GtkWidget *widget, Chmsee *self)
 /* internal functions */
 
 static void
-chmsee_quit(Chmsee *self)
+on_quit(GtkAction *action, Chmsee *self)
 {
         g_message("Chmsee >>> quit");
 
@@ -687,12 +706,13 @@ chmsee_quit(Chmsee *self)
         cs_html_gecko_shutdown_system();
         priv->config->hpaned_pos = cs_book_get_hpaned_position(CS_BOOK (priv->book));
         gtk_widget_destroy(GTK_WIDGET (self));
+
         gtk_main_quit();
 }
 
 static void
-chmsee_drag_data_received (GtkWidget *widget, GdkDragContext *context, gint x, gint y,
-                           GtkSelectionData *selection_data, guint info, guint time)
+chmsee_drag_data_received(GtkWidget *widget, GdkDragContext *context, gint x, gint y,
+                          GtkSelectionData *selection_data, guint info, guint time)
 {
         gchar  **uris;
         gint     i = 0;
@@ -745,6 +765,8 @@ populate_windows(Chmsee *self)
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "Home"), FALSE);
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "Back"), FALSE);
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "Forward"), FALSE);
+        gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "Prev"), FALSE);
+        gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "Next"), FALSE);
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "SidePane"), FALSE);
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "ZoomIn"), FALSE);
         gtk_action_set_sensitive(gtk_action_group_get_action(action_group, "ZoomOut"), FALSE);
@@ -922,8 +944,6 @@ chmsee_new(CsConfig *config)
 
         priv->config = config;
 
-        cs_html_gecko_set_default_lang(config->lang);
-
         if (config->pos_x >= 0 && config->pos_y >= 0)
                 gtk_window_move(GTK_WINDOW (self), config->pos_x, config->pos_y);
 
@@ -987,7 +1007,7 @@ chmsee_open_file(Chmsee *self, const gchar *filename)
                 gtk_window_set_title(GTK_WINDOW (self), window_title);
                 g_free(window_title);
 
-                /* record last opened file directory */
+                /* record last opened file */
                 if (priv->config->last_file)
                         g_free(priv->config->last_file);
 
@@ -1038,20 +1058,6 @@ chmsee_open_file(Chmsee *self, const gchar *filename)
         }
 }
 
-int
-chmsee_get_lang(Chmsee *self)
-{
-        return CHMSEE_GET_PRIVATE (self)->config->lang;
-}
-
-void
-chmsee_set_lang(Chmsee *self, int lang)
-{
-        ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
-        priv->config->lang = lang;
-        cs_html_gecko_set_default_lang(lang);
-}
-
 gboolean
 chmsee_get_startup_lastfile(Chmsee *self)
 {
@@ -1065,6 +1071,19 @@ chmsee_set_startup_lastfile(Chmsee *self, gboolean state)
         CHMSEE_GET_PRIVATE (self)->config->startup_lastfile = state;
 }
 
+gboolean
+chmsee_get_splash(Chmsee *self)
+{
+        return CHMSEE_GET_PRIVATE (self)->config->splash;
+}
+
+void
+chmsee_set_splash(Chmsee *self, gboolean state)
+{
+        g_debug("Chmsee >>> set splash = %d", state);
+        CHMSEE_GET_PRIVATE (self)->config->splash = state;
+}
+
 const gchar *
 chmsee_get_variable_font(Chmsee *self)
 {
@@ -1096,6 +1115,21 @@ chmsee_set_fixed_font(Chmsee *self, const gchar *font_name)
         cs_chmfile_set_fixed_font(priv->chmfile, font_name);
 }
 
+const gchar *
+chmsee_get_charset(Chmsee *self)
+{
+        ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
+        return cs_chmfile_get_charset(priv->chmfile);
+}
+
+void
+chmsee_set_charset(Chmsee *self, const gchar *charset)
+{
+        ChmseePrivate *priv = CHMSEE_GET_PRIVATE (self);
+        cs_chmfile_set_charset(priv->chmfile, charset);
+        cs_book_reload_current_page(CS_BOOK (priv->book));
+}
+
 gboolean
 chmsee_has_book(Chmsee *self)
 {
