@@ -61,7 +61,6 @@ struct _CsConfig {
         gchar   *bookshelf;
         gchar   *last_file;
 
-        gint     lang;
         gint     pos_x;
         gint     pos_y;
         gint     height;
@@ -69,6 +68,7 @@ struct _CsConfig {
         gint     hpaned_pos;
         gboolean fullscreen;
         gboolean startup_lastfile;
+        gboolean splash;
 };
 
 GType        chmsee_get_type(void);
@@ -76,15 +76,15 @@ Chmsee      *chmsee_new(CsConfig *);
 void         chmsee_open_file(Chmsee *, const gchar *);
 void         chmsee_close_book(Chmsee *);
 
-int          chmsee_get_lang(Chmsee *);
-void         chmsee_set_lang(Chmsee *, int);
-
 const gchar *chmsee_get_variable_font(Chmsee *);
 void         chmsee_set_variable_font(Chmsee *, const gchar *);
 
 const gchar *chmsee_get_fixed_font(Chmsee *);
 void         chmsee_set_fixed_font(Chmsee *, const gchar *);
 
+const gchar *chmsee_get_charset(Chmsee *);
+void         chmsee_set_charset(Chmsee *, const gchar *);
+
 gboolean     chmsee_get_startup_lastfile(Chmsee *);
 void         chmsee_set_startup_lastfile(Chmsee *, gboolean);
 
