@@ -95,14 +95,13 @@ static void html_open_new_tab_cb(CsHtmlGecko *, const gchar *, CsBook *);
 static void html_link_message_cb(CsHtmlGecko *, const gchar *, CsBook *);
 
 static void on_tab_close(GtkWidget *, CsBook *);
-static void on_copy(GtkWidget *, CsBook *);
-static void on_copy_page_location(GtkWidget *, CsBook *);
-static void on_select_all(GtkWidget *, CsBook *);
-static void on_back(GtkWidget *, CsBook *);
-static void on_forward(GtkWidget *, CsBook *);
-static void on_zoom_in(GtkWidget *, CsBook *);
-static void on_context_new_tab(GtkWidget *, CsBook *);
-static void on_context_copy_link(GtkWidget *, CsBook *);
+static void on_copy(GtkAction *, CsBook *);
+static void on_copy_page_location(GtkAction *, CsBook *);
+static void on_select_all(GtkAction *, CsBook *);
+static void on_back(GtkAction *, CsBook *);
+static void on_forward(GtkAction *, CsBook *);
+static void on_context_new_tab(GtkAction *, CsBook *);
+static void on_context_copy_link(GtkAction *, CsBook *);
 static void on_findbar_hide(GtkWidget *, CsBook *);
 static void on_findbar_back(GtkWidget *, CsBook *);
 static void on_findbar_forward(GtkWidget *, CsBook *);
@@ -114,6 +113,8 @@ static void update_tab_title(CsBook *, CsHtmlGecko *, const gchar *);
 static void update_tab_label_state(CsBook *);
 static void set_context_menu_link(CsBook *, const gchar *);
 static void find_text(GtkWidget *, CsBook *, gboolean);
+static gint uri_compare(gconstpointer, gconstpointer);
+static gchar *get_short_uri(CsChmfile *, const gchar *);
 
 static const GtkActionEntry entries[] = {
         { "Copy", GTK_STOCK_COPY, N_("_Copy"), "<control>C", NULL, G_CALLBACK(on_copy)},
@@ -123,7 +124,6 @@ static const GtkActionEntry entries[] = {
         { "CopyLinkLocation", NULL, N_("_Copy Link Location"), NULL, NULL, G_CALLBACK(on_context_copy_link)},
         { "SelectAll", NULL, N_("Select _All"), NULL, NULL, G_CALLBACK(on_select_all)},
         { "CopyPageLocation", NULL, N_("Copy Page _Location"), NULL, NULL, G_CALLBACK(on_copy_page_location)},
-        { "OnKeyboardControlEqual", NULL, NULL, "<control>equal", NULL, G_CALLBACK(on_zoom_in)},
 };
 
 /* Radio items */
@@ -147,7 +147,6 @@ static const char *ui_description =
         "    <menuitem action='SelectAll'/>"
         "    <menuitem action='CopyPageLocation'/>"
         "  </popup>"
-        "  <accelerator action='OnKeyboardControlEqual'/>"
         "</ui>";
 
 
@@ -369,15 +368,11 @@ link_selected_cb(GtkWidget *widget, Link *link, CsBook *self)
         char *scheme = g_uri_parse_scheme(link->uri);
         if (scheme && g_strcmp0(scheme, "file")) {
                 g_message("%s is unsupported protocol.", scheme);
-                gchar *message = g_strdup_printf("URI %s has unsupported protocol: %s", link->uri, scheme);
+                gchar *message = g_strdup_printf("URI %s with unsupported protocol: %s", link->uri, scheme);
                 update_book_message(self, message);
                 g_free(message);
         } else {
-                cs_book_load_url(self, link->uri);
-
-                CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-                if (widget != priv->toc_page)
-                        cs_toc_select_uri(CS_TOC (priv->toc_page), link->uri);
+                cs_book_load_url(self, link->uri, FALSE);
         }
         g_free(scheme);
 }
@@ -394,7 +389,7 @@ html_notebook_switch_page_cb(GtkNotebook *notebook, GtkNotebookPage *page, guint
         if (new_page) {
                 g_debug("CS_BOOK >>> switch page callback, set active_html = %p", priv->active_html);
                 priv->active_html = CS_HTML_GECKO (new_page);
-                cs_html_gecko_reload(priv->active_html);
+                cs_book_reload_current_page(self);
         }
 
         g_signal_emit(self, signals[HTML_CHANGED], 0, self);
@@ -406,6 +401,24 @@ html_location_changed_cb(CsHtmlGecko *html, const gchar *location, CsBook *self)
         g_debug("CS_BOOK >>> html location changed cb: %s", location);
 
         g_signal_emit(self, signals[HTML_CHANGED], 0, self);
+
+        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
+
+        char *scheme = g_uri_parse_scheme(location);
+
+        if (!g_strcmp0(scheme, "file") && priv->toc_page) {
+                gchar *real_uri = get_real_uri(location);
+                gchar *filename = g_filename_from_uri(real_uri, NULL, NULL);
+
+                gchar *uri = get_short_uri(priv->model, filename);
+                gchar *toc_uri = g_strdup_printf("%s%s", uri, location + strlen(real_uri));
+
+                cs_toc_sync(CS_TOC (priv->toc_page), toc_uri);
+
+                g_free(real_uri);
+                g_free(filename);
+                g_free(toc_uri);
+        }
 }
 
 static gboolean
@@ -420,37 +433,24 @@ html_open_uri_cb(CsHtmlGecko *html, const gchar *full_uri, CsBook *self)
         char *scheme = g_uri_parse_scheme(full_uri);
 
         if (scheme) {
+                const gchar *bookfolder = cs_chmfile_get_bookfolder(priv->model);
+
                 if (!g_strcmp0(scheme, "file")) {
                         /* DND chmfile check */
                         if (g_str_has_suffix(full_uri, ".chm") || g_str_has_suffix(full_uri, ".CHM")) {
                                 g_debug("CS_BOOK >>> open chm file = %s", full_uri);
                                 g_signal_emit(self, signals[MODEL_CHANGED], 0, NULL, full_uri);
-                        } else {
-                                const gchar *bookfolder = cs_chmfile_get_bookfolder(priv->model);
-                                gchar *uri = g_strrstr(full_uri, bookfolder);
-                                uri = uri + strlen(bookfolder);
+                        } else if (g_strrstr(full_uri, bookfolder)) {
+                                gchar *uri = get_short_uri(priv->model, full_uri);
+
                                 g_debug("CS_BOOK >>> html_open_uri call load url = %s", uri);
-                                cs_book_load_url(self, uri);
-
-                                if (priv->toc_page) {
-                                        gchar *real_uri = get_real_uri(full_uri);
-                                        gchar *filename = g_filename_from_uri(real_uri, NULL, NULL);
-
-                                        g_debug("CS_BOOK >>> html_open_uri filename = %s", filename);
-                                        uri = g_strrstr(filename, bookfolder);
-                                        uri = uri + strlen(bookfolder);
-                                        gchar *toc_uri = g_strdup_printf("%s%s", uri, full_uri+strlen(real_uri));
-                                        g_debug("CS_BOOK >>> html_open_uri toc_uri= %s", toc_uri);
-                                        cs_toc_select_uri(CS_TOC (priv->toc_page), toc_uri);
-                                        g_free(real_uri);
-                                        g_free(filename);
-                                        g_free(toc_uri);
-                                }
+                                cs_book_load_url(self, uri, TRUE);
                         }
                 } else if (!g_strcmp0(scheme, "about") || !g_strcmp0(scheme, "jar")) {
                         return FALSE;
                 }
         }
+
         return TRUE;
 }
 
@@ -468,16 +468,15 @@ html_title_changed_cb(CsHtmlGecko *html, const gchar *title, CsBook *self)
 
         update_tab_title(self, html, label_text);
 
-        /* sync bookmarks title entry */
+        /* update bookmarks title entry */
         gchar *location = cs_html_gecko_get_location(html);
-        g_debug("CS_BOOK >>> html title changed cb location = %s", location);
 
         if (location != NULL && strlen(location)) {
                 gboolean about = g_str_has_prefix(location, "about:");
                 if (!about) {
-                        const gchar *bookfolder = cs_chmfile_get_bookfolder(priv->model);
-                        gchar *uri = g_strrstr(location, bookfolder) + strlen(bookfolder);
+                        gchar *uri = get_short_uri(priv->model, location);
                         Link *link = link_new(LINK_TYPE_PAGE, label_text, uri);
+
                         cs_bookmarks_set_current_link(CS_BOOKMARKS (priv->bookmarks_page), link);
                         link_free(link);
                 }
@@ -558,15 +557,13 @@ on_tab_close(GtkWidget *widget, CsBook *self)
 }
 
 static void
-on_copy(GtkWidget *widget, CsBook *self)
+on_copy(GtkAction *action, CsBook *self)
 {
-        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-
-        cs_html_gecko_copy_selection(priv->active_html);
+        cs_book_copy(self);
 }
 
 static void
-on_copy_page_location(GtkWidget *widget, CsBook *self)
+on_copy_page_location(GtkAction *action, CsBook *self)
 {
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
 
@@ -585,35 +582,25 @@ on_copy_page_location(GtkWidget *widget, CsBook *self)
 }
 
 static void
-on_select_all(GtkWidget *widget, CsBook *self)
+on_select_all(GtkAction *action, CsBook *self)
 {
-        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-        cs_html_gecko_select_all(priv->active_html);
+        cs_book_select_all(self);
 }
 
 static void
-on_back(GtkWidget *widget, CsBook *self)
+on_back(GtkAction *action, CsBook *self)
 {
-        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-        cs_html_gecko_go_back(priv->active_html);
+        cs_book_go_back(self);
 }
 
 static void
-on_forward(GtkWidget *widget, CsBook *self)
+on_forward(GtkAction *action, CsBook *self)
 {
-        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-        cs_html_gecko_go_forward(priv->active_html);
+        cs_book_go_forward(self);
 }
 
 static void
-on_zoom_in(GtkWidget *widget, CsBook *self)
-{
-        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
-        cs_html_gecko_increase_size(priv->active_html);
-}
-
-static void
-on_context_new_tab(GtkWidget *widget, CsBook *self)
+on_context_new_tab(GtkAction *action, CsBook *self)
 {
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
         g_debug("CS_BOOK >>> On context open new tab: %s", priv->context_menu_link);
@@ -624,7 +611,7 @@ on_context_new_tab(GtkWidget *widget, CsBook *self)
 }
 
 static void
-on_context_copy_link(GtkWidget *widget, CsBook *self)
+on_context_copy_link(GtkAction *action, CsBook *self)
 {
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
         g_debug("CS_BOOK >>> On context copy link: %s", priv->context_menu_link);
@@ -850,6 +837,29 @@ update_tab_label_state(CsBook *self)
         }
 }
 
+static gint
+uri_compare(gconstpointer a, gconstpointer b)
+{
+        return ncase_compare_utf8_string(((Link *)a)->uri, (gchar *)b);
+}
+
+static gchar *
+get_short_uri(CsChmfile *chmfile, const gchar *uri)
+{
+        const gchar *bookfolder = cs_chmfile_get_bookfolder(chmfile);
+        gchar *short_uri = g_strrstr(uri, bookfolder);
+
+        if (!short_uri)
+                short_uri = uri;
+        else
+                short_uri = short_uri + strlen(bookfolder);
+
+        if (short_uri[0] == '/')
+                short_uri = short_uri + 1;
+
+        return short_uri;
+}
+
 /* External functions*/
 
 GtkWidget *
@@ -956,52 +966,67 @@ cs_book_set_model(CsBook *self, CsChmfile *model)
         gtk_notebook_set_current_page(GTK_NOTEBOOK (priv->html_notebook), cur_page);
         gtk_widget_show_all(priv->html_notebook);
 
-        cs_book_homepage(self);
+        const gchar *page = cs_chmfile_get_page(model);
+        if (page) {
+                cs_book_load_url(self, page, TRUE);
+        } else {
+                cs_book_homepage(self);
+        }
 
         g_signal_emit(self, signals[MODEL_CHANGED], 0, model, NULL);
 }
 
 void
-cs_book_load_url(CsBook *self, const gchar *uri)
+cs_book_load_url(CsBook *self, const gchar *uri, gboolean force_reload)
 {
-        g_debug("CS_BOOK >>> cs_book_load_url %s", uri);
+        g_debug("CS_BOOK >>> load uri %s", uri);
         g_return_if_fail(IS_CS_BOOK (self));
 
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
         gchar *full_uri;
 
+        /* assemble input uri with bookfolder */
         if (strlen(uri)) {
                 if (uri[0] == '/')
                         full_uri = g_strdup_printf("file://%s%s", cs_chmfile_get_bookfolder(priv->model), uri);
                 else
                         full_uri = g_strdup_printf("file://%s/%s", cs_chmfile_get_bookfolder(priv->model), uri);
         } else {
-                full_uri = cs_html_gecko_get_location(priv->active_html);
+                return; /* empty page do nothing*/
         }
 
         /* check file exist */
         gchar *real_uri = get_real_uri(full_uri);
         gchar *filename = g_filename_from_uri(real_uri, NULL, NULL);
+        gboolean file_exist = FALSE;
 
-        gboolean has_file = FALSE;
         if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
-                has_file = TRUE;
+                file_exist = TRUE;
         } else {
+                /* search again with case insensitive name */
                 gchar *found = file_exist_ncase(filename);
                 if (found) {
                         g_free(full_uri);
 
                         full_uri = g_strdup_printf("file://%s%s", found, full_uri+strlen(real_uri));
                         g_free(found);
-                        has_file = TRUE;
+                        file_exist = TRUE;
                 }
         }
 
-        if (has_file) {
-                g_debug("CS_BOOK >>> cs_book_load_url html = %p, full_uri = %s", priv->active_html, full_uri);
-                g_signal_handlers_block_by_func(priv->active_html, html_open_uri_cb, self);
-                cs_html_gecko_load_url(priv->active_html, full_uri);
-                g_signal_handlers_unblock_by_func(priv->active_html, html_open_uri_cb, self);
+        if (file_exist) {
+                gchar *location = cs_html_gecko_get_location(priv->active_html);
+                if (force_reload || g_strcmp0(full_uri, location)) {
+                        /* set user specified charset */
+                        const gchar *charset = cs_chmfile_get_charset(CS_CHMFILE (priv->model));
+                        if (charset && strlen(charset))
+                                cs_html_gecko_set_charset(priv->active_html, charset);
+
+                        g_signal_handlers_block_by_func(priv->active_html, html_open_uri_cb, self);
+                        cs_html_gecko_load_url(priv->active_html, full_uri);
+                        g_signal_handlers_unblock_by_func(priv->active_html, html_open_uri_cb, self);
+                }
+                g_free(location);
         } else {
                 GtkWidget *msg_dialog;
 
@@ -1014,7 +1039,7 @@ cs_book_load_url(CsBook *self, const gchar *uri)
                 gtk_dialog_run(GTK_DIALOG (msg_dialog));
                 gtk_widget_destroy(msg_dialog);
 
-                g_message("CS_BOOK >>> cannot found target file = %s", filename);
+                g_message("CS_BOOK >>> cannot find target file = %s", filename);
         }
 
         g_free(full_uri);
@@ -1034,25 +1059,17 @@ cs_book_new_tab_with_fulluri(CsBook *self, const gchar *full_uri)
 
         char *scheme = g_uri_parse_scheme(full_uri);
         if (scheme && g_strcmp0(scheme, "file")) {
-                gchar *message = g_strdup_printf("URI %s has unsupported protocol: %s", full_uri, scheme);
+                gchar *message = g_strdup_printf("URI %s with unsupported protocol: %s", full_uri, scheme);
                 g_message("%s", message);
                 update_book_message(self, message);
                 g_free(message);
         } else {
-                const gchar *bookfolder = cs_chmfile_get_bookfolder(priv->model);
-                g_debug("CS_BOOK >>> new tab with full url, get bookfolder = %s", bookfolder);
-
-                gchar *uri = g_strrstr(full_uri, bookfolder);
-                uri = uri + strlen(bookfolder);
-
                 gint page_num = new_html_tab(self);
                 gtk_notebook_set_current_page(GTK_NOTEBOOK (priv->html_notebook), page_num);
                 update_tab_label_state(self);
 
-                cs_book_load_url(self, uri);
-
-                if (priv->toc_page)
-                        cs_toc_select_uri(CS_TOC (priv->toc_page), uri);
+                gchar *uri = get_short_uri(priv->model, full_uri);
+                cs_book_load_url(self, uri, TRUE);
         }
         g_free(scheme);
 }
@@ -1097,8 +1114,12 @@ cs_book_reload_current_page(CsBook *self)
         g_debug("CS_BOOK >>> Reload current page");
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
 
-        if (priv->model)
+        if (priv->model) {
+                const gchar *charset = cs_chmfile_get_charset(CS_CHMFILE (priv->model));
+                if (charset && strlen(charset))
+                        cs_html_gecko_set_charset(priv->active_html, charset);
                 cs_html_gecko_reload(priv->active_html);
+        }
 }
 
 void
@@ -1110,10 +1131,7 @@ cs_book_homepage(CsBook *self)
         const gchar *homepage = cs_chmfile_get_homepage(priv->model);
 
         if (homepage) {
-                cs_book_load_url(self, homepage);
-
-                if (priv->toc_page)
-                        cs_toc_select_uri(CS_TOC (priv->toc_page), homepage);
+                cs_book_load_url(self, homepage, FALSE);
         }
 }
 
@@ -1151,6 +1169,7 @@ cs_book_can_go_forward(CsBook *self)
 void
 cs_book_go_back(CsBook *self)
 {
+        g_debug("CS_BOOK >>> go back");
         g_return_if_fail(IS_CS_BOOK (self));
         CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
         cs_html_gecko_go_back(priv->active_html);
@@ -1164,6 +1183,47 @@ cs_book_go_forward(CsBook *self)
         cs_html_gecko_go_forward(priv->active_html);
 }
 
+void
+cs_book_go_prev(CsBook *self)
+{
+        g_debug("CS_BOOK >>> go prev");
+        g_return_if_fail(IS_CS_BOOK (self));
+
+        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
+
+        GList *toc_list = cs_chmfile_get_toc_list(priv->model);
+        gchar *location = cs_html_gecko_get_location(priv->active_html);
+        gchar *short_uri = get_short_uri(priv->model, location);
+        GList *current = g_list_find_custom(toc_list, short_uri, uri_compare);
+
+        g_free(location);
+
+        if (current && current->prev) {
+                gchar *uri = ((Link *)current->prev->data)->uri;
+                cs_book_load_url(self, uri, FALSE);
+        }
+}
+
+void
+cs_book_go_next(CsBook *self)
+{
+        g_debug("CS_BOOK >>> go next");
+        g_return_if_fail(IS_CS_BOOK (self));
+
+        CsBookPrivate *priv = CS_BOOK_GET_PRIVATE(self);
+
+        GList *toc_list = cs_chmfile_get_toc_list(priv->model);
+        gchar *location = cs_html_gecko_get_location(priv->active_html);
+        gchar *short_uri = get_short_uri(priv->model, location);
+        GList *current = g_list_find_custom(toc_list, short_uri, uri_compare);
+        g_free(location);
+
+        if (current && current->next) {
+                gchar *uri = ((Link *)current->next->data)->uri;
+                cs_book_load_url(self, uri, FALSE);
+        }
+}
+
 void
 cs_book_zoom_in(CsBook *self)
 {
