@@ -98,7 +98,6 @@ load_config()
         if (!g_file_test(config->bookshelf, G_FILE_TEST_IS_DIR))
                 mkdir(config->bookshelf, 0755);
 
-        config->lang       = 0;
         config->last_file  = NULL;
         config->pos_x      = -100;
         config->pos_y      = -100;
@@ -107,6 +106,7 @@ load_config()
         config->hpaned_pos = 200;
         config->fullscreen       = FALSE;
         config->startup_lastfile = FALSE;
+        config->splash = TRUE;
 
         gchar *config_file = g_build_filename(config->home, CHMSEE_CONFIG_FILE, NULL);
         g_debug("Main >>> chmsee config file path = %s", config_file);
@@ -115,7 +115,6 @@ load_config()
 
         GKeyFile *keyfile = g_key_file_new();
         gboolean rv = g_key_file_load_from_file(keyfile, config_file, G_KEY_FILE_NONE, NULL);
-        g_debug("Main >>> chmsee load config file return = %d", rv);
 
         if (!rv) {
                 convert_old_config_file(config_file, "[ChmSee]\n");
@@ -124,7 +123,6 @@ load_config()
         rv = g_key_file_load_from_file(keyfile, config_file, G_KEY_FILE_NONE, NULL);
 
         if (rv) {
-                config->lang      = g_key_file_get_integer(keyfile, "ChmSee", "LANG", NULL);
                 config->last_file = g_key_file_get_string (keyfile, "ChmSee", "LAST_FILE", NULL);
 
                 config->pos_x      = g_key_file_get_integer(keyfile, "ChmSee", "POS_X", NULL);
@@ -134,6 +132,7 @@ load_config()
                 config->hpaned_pos = g_key_file_get_integer(keyfile, "ChmSee", "HPANED_POSITION", NULL);
                 config->fullscreen       = g_key_file_get_boolean(keyfile, "ChmSee", "FULLSCREEN", NULL);
                 config->startup_lastfile = g_key_file_get_boolean(keyfile, "ChmSee", "STARTUP_LASTFILE", NULL);
+                config->splash           = g_key_file_get_boolean(keyfile, "ChmSee", "SPLASH", NULL);
 
                 if (!config->hpaned_pos)
                         config->hpaned_pos = 200;
@@ -154,7 +153,6 @@ save_config(CsConfig *config)
         gchar *config_file = g_build_filename(config->home, CHMSEE_CONFIG_FILE, NULL);
 
         GKeyFile *keyfile = g_key_file_new();
-        g_key_file_set_integer(keyfile, "ChmSee", "LANG", config->lang);
 
         if (config->last_file)
                 g_key_file_set_string (keyfile, "ChmSee", "LAST_FILE", config->last_file);
@@ -166,6 +164,7 @@ save_config(CsConfig *config)
         g_key_file_set_integer(keyfile, "ChmSee", "HPANED_POSITION", config->hpaned_pos);
         g_key_file_set_boolean(keyfile, "ChmSee", "FULLSCREEN", config->fullscreen);
         g_key_file_set_boolean(keyfile, "ChmSee", "STARTUP_LASTFILE", config->startup_lastfile);
+        g_key_file_set_boolean(keyfile, "ChmSee", "SPLASH", config->splash);
 
         gchar *contents = g_key_file_to_data(keyfile, &length, NULL);
         g_file_set_contents(config_file, contents, length, NULL);
@@ -193,20 +192,22 @@ main(int argc, char *argv[])
         if (!g_thread_supported())
                 g_thread_init(NULL);
 
+        gdk_threads_init();
+
         GOptionEntry options[] = {
                 {"version", 0,
                  0, G_OPTION_ARG_NONE, &option_version,
-                 _("Display the version and exit"),
+                 _("Display ChmSee version"),
                  NULL
                 },
                 {"verbose", 'v',
                  G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK, (void*)callback_verbose,
-                 _("be verbose, repeat 3 times to get all info"),
+                 _("Be verbose, repeat 3 times to get all infomation"),
                  NULL
                 },
                 {"quiet", 'q',
                  G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK, (void*)callback_quiet,
-                 _("be quiet, repeat 2 times to disable all info"),
+                 _("Be quiet, repeat 2 times to disable all infomation"),
                  NULL
                 },
                 /* {"bookshelf", 0, */
@@ -221,7 +222,7 @@ main(int argc, char *argv[])
                 "[chmfile]\n"
                 "\n"
                 "GTK+ based CHM file viewer\n"
-                "Example: chmsee FreeBSD_Handbook.chm";
+                "Example: chmsee Handbook.chm::toc.html";
 
         if (!gtk_init_with_args(&argc, &argv, params, options, GETTEXT_PACKAGE, &error)) {
                 g_printerr("%s\n", error->message);
@@ -248,19 +249,22 @@ main(int argc, char *argv[])
         bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
         textdomain(GETTEXT_PACKAGE);
 
-        /* Show splash screen */
-        startup_popup_new();
-
         CsConfig *config = load_config();
 
         /* if (bookshelf != NULL) { */
         /*         config->bookshelf = g_strdup(bookshelf); */
         /* } */
 
+        gdk_threads_enter();
+
+        /* Show splash screen */
+        if(config->splash)
+                startup_popup_new();
+
         Chmsee *chmsee = chmsee_new(config);
 
         if (chmsee == NULL) {
-                g_warning("Create chmsee main window failed!");
+                g_warning("Creating chmsee main window failed!");
                 return 1;
         }
 
@@ -271,6 +275,8 @@ main(int argc, char *argv[])
 
         gtk_main();
 
+        gdk_threads_leave();
+
         save_config(config);
 
         return 0;
