@@ -29,6 +29,8 @@
 #include <sys/stat.h>
 #include <gcrypt.h>
 #include <chm_lib.h>
+#include <gtk/gtkwindow.h>
+#include <gtk/gtk.h>
 
 #include "chmfile.h"
 #include "utils.h"
@@ -40,19 +42,22 @@ typedef struct _CsChmfilePrivate CsChmfilePrivate;
 
 struct _CsChmfilePrivate {
         GNode       *toc_tree;
+        GList       *toc_list;
         GList       *index_list;
         GList       *bookmarks_list;
 
         gchar       *hhc;
         gchar       *hhk;
         gchar       *bookfolder;     /* the folder CHM file extracted to */
-        gchar       *filename;       /* opened CHM file name */
+        gchar       *chm;            /* opened CHM file name */
+        gchar       *page;           /* :: specified page */
 
         gchar       *bookname;
         gchar       *homepage;
-        gchar       *encoding;
+        gchar       *encoding;       /* retrieved from chmfile */
         gchar       *variable_font;
         gchar       *fixed_font;
+        gchar       *charset;        /* user specified on setup window */
 };
 
 struct extract_context
@@ -62,6 +67,13 @@ struct extract_context
         const char *hhk_file;
 };
 
+struct extract_data
+{
+        const gchar *filename;
+        const gchar *bookfolder;
+        int          rval;
+};
+
 #define CS_CHMFILE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CS_TYPE_CHMFILE, CsChmfilePrivate))
 
 #define UINT16ARRAY(x) ((unsigned char)(x)[0] | ((u_int16_t)(x)[1] << 8))
@@ -86,11 +98,13 @@ static void load_bookinfo(CsChmfile *);
 static void save_bookinfo(CsChmfile *);
 static void extract_post_file_write(const gchar *);
 
-static GNode *parse_hhc_file(const gchar *, const gchar*);
-static void parse_hhk_file(CsChmfile *, const gchar *, const gchar*);
-static void tree_to_list_callback(GNode *, gpointer);
+static GList *convert_node_to_list(GNode *);
+static gboolean tree_to_list_callback(GNode *, gpointer);
 static void free_list_data(gpointer, gpointer);
-static gchar *check_file_ncase(CsChmfile *, gchar *);
+static gchar *check_file_ncase(CsChmfile *, const gchar *);
+
+static void extract_file(struct extract_data *);
+static void parse_filename(CsChmfile *, const gchar *);
 
 /* GObject functions */
 
@@ -118,7 +132,8 @@ cs_chmfile_init(CsChmfile *self)
         priv->hhc            = NULL;
         priv->hhk            = NULL;
         priv->bookfolder     = NULL;
-        priv->filename       = NULL;
+        priv->chm            = NULL;
+        priv->page           = NULL;
 
         priv->bookname       = NULL;
         priv->homepage       = NULL;
@@ -126,6 +141,7 @@ cs_chmfile_init(CsChmfile *self)
         priv->encoding       = g_strdup("UTF-8");
         priv->variable_font  = g_strdup("Sans 12");
         priv->fixed_font     = g_strdup("Monospace 12");
+        priv->charset        = g_strdup("");
 }
 
 static void
@@ -141,13 +157,15 @@ cs_chmfile_finalize(GObject *object)
         g_free(priv->hhc);
         g_free(priv->hhk);
         g_free(priv->bookfolder);
-        g_free(priv->filename);
+        g_free(priv->chm);
+        g_free(priv->page);
 
         g_free(priv->bookname);
         g_free(priv->homepage);
         g_free(priv->encoding);
         g_free(priv->variable_font);
         g_free(priv->fixed_font);
+        g_free(priv->charset);
 
         if (priv->index_list) {
                 g_list_foreach(priv->index_list, (GFunc)free_list_data, NULL);
@@ -168,6 +186,19 @@ cs_chmfile_finalize(GObject *object)
 
 /* Internal functions */
 
+static void
+extract_file(struct extract_data *data)
+{
+        gboolean rval = extract_chm(data->filename, data->bookfolder);
+
+        gdk_threads_enter();
+        if (rval)
+                data->rval = 0;
+        else
+                data->rval = 1;
+        gdk_threads_leave();
+}
+
 static int
 dir_exists(const char *path)
 {
@@ -509,10 +540,10 @@ chmfile_file_info(CsChmfile *self)
 {
         CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
 
-        struct chmFile *cfd = chm_open(priv->filename);
+        struct chmFile *cfd = chm_open(priv->chm);
 
         if (cfd == NULL) {
-                g_error(_("Can not open chm file %s."), priv->filename);
+                g_error(_("Can not open chm file %s."), priv->chm);
                 return;
         }
 
@@ -520,43 +551,60 @@ chmfile_file_info(CsChmfile *self)
         chmfile_windows_info(cfd, self);
 
         if (priv->hhc != NULL) {
-                check_file_ncase(self, priv->hhc);
+                gchar *new_hhc = check_file_ncase(self, priv->hhc);
+
+                if (new_hhc) {
+                        g_free(priv->hhc);
+                        priv->hhc = new_hhc;
+                }
         }
 
         if (priv->hhk != NULL) {
-                check_file_ncase(self, priv->hhk);
+                gchar *new_hhk = check_file_ncase(self, priv->hhk);
+
+                if (new_hhk) {
+                        g_free(priv->hhk);
+                        priv->hhk = new_hhk;
+                }
         }
 
         /* convert encoding to UTF-8 */
         if (priv->encoding != NULL) {
                 if (priv->bookname) {
+                        g_debug("CS_CHMFILE >>> priv->bookname = %s", priv->bookname);
                         gchar *bookname_utf8 = convert_string_to_utf8(priv->bookname, priv->encoding);
-                        g_debug("CS_CHMFILE >>> priv->bookname = %s, bookname_utf8 = %s", priv->bookname, bookname_utf8);
+                        g_debug("CS_CHMFILE >>> bookname_utf8 = %s", bookname_utf8);
                         g_free(priv->bookname);
                         priv->bookname = bookname_utf8;
                 }
 
                 if (priv->hhc) {
+                        g_debug("CS_CHMFILE >>> priv->hhc = %s", priv->hhc);
                         gchar *hhc_utf8 = convert_filename_to_utf8(priv->hhc, priv->encoding);
                         g_free(priv->hhc);
                         priv->hhc = hhc_utf8;
                 }
 
                 if (priv->hhk) {
+                        g_debug("CS_CHMFILE >>> priv->hhk = %s", priv->hhk);
                         gchar *hhk_utf8 = convert_filename_to_utf8(priv->hhk, priv->encoding);
                         g_free(priv->hhk);
                         priv->hhk = hhk_utf8;
                 }
 
                 if (priv->homepage) {
+                        g_debug("CS_CHMFILE >>> priv->homepage = %s", priv->homepage);
                         gchar *homepage_utf8 = convert_filename_to_utf8(priv->homepage, priv->encoding);
                         g_free(priv->homepage);
                         priv->homepage = homepage_utf8;
                 }
         }
 
-        if (priv->bookname == NULL)
-                priv->bookname = g_path_get_basename(priv->filename);
+        if (priv->bookname == NULL) {
+                gchar *bookname = g_path_get_basename(priv->chm);
+                priv->bookname = g_strdup(bookname);
+                g_free(bookname);
+        }
 
         chm_close(cfd);
 }
@@ -727,33 +775,27 @@ chmfile_system_info(struct chmFile *cfd, CsChmfile *self)
         }
 }
 
-static GNode *
-parse_hhc_file(const gchar *file, const gchar *encoding)
+static GList *
+convert_node_to_list(GNode *tree)
 {
-        g_debug("CS_CHMFILE >>> parse hhc file = %s, encoding = %s", file, encoding);
-        return cs_parse_file(file, encoding);
+        GList root;
+        root.data = NULL;
+        g_node_traverse(tree, G_PRE_ORDER, G_TRAVERSE_ALL, -1, tree_to_list_callback, &root);
+        return (GList *)root.data;
 }
 
-static void
-parse_hhk_file(CsChmfile *self, const gchar *file, const gchar *encoding)
-{
-        CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
-
-        g_debug("CHMFILE >>> parse hhk file = %s, encoding = %s", file, encoding);
-
-        GNode *tree = cs_parse_file(file, encoding);
-        g_debug("CHMFILE >>> parse hhk file tree = %p", tree);
-
-        /* convert GNode to GList */
-        g_node_children_foreach(tree, G_TRAVERSE_LEAVES, tree_to_list_callback, priv);
-        g_node_destroy(tree);
-}
-
-static void
+static gboolean
 tree_to_list_callback(GNode *node, gpointer data)
 {
-        CsChmfilePrivate *priv = (CsChmfilePrivate *)data;
-        priv->index_list = g_list_append(priv->index_list, node->data);
+        GList *root = (GList *)data;
+
+        if (node->parent) {
+                Link *link = (Link *)node->data;
+                if (g_ascii_strcasecmp(CHMSEE_NO_LINK, link->uri) && strlen(link->uri))
+                        root->data = g_list_append((GList *)root->data, node->data);
+        }
+
+        return FALSE;
 }
 
 static void
@@ -790,6 +832,7 @@ load_bookinfo(CsChmfile *self)
                 priv->encoding      = g_key_file_get_string(keyfile, "Bookinfo", "encoding", NULL);
                 priv->variable_font = g_key_file_get_string(keyfile, "Bookinfo", "variable_font", NULL);
                 priv->fixed_font    = g_key_file_get_string(keyfile, "Bookinfo", "fixed_font", NULL);
+                priv->charset       = g_key_file_get_string(keyfile, "Bookinfo", "charset", NULL);
         }
 
         g_key_file_free(keyfile);
@@ -818,6 +861,7 @@ save_bookinfo(CsChmfile *self)
         g_key_file_set_string(keyfile, "Bookinfo", "encoding", priv->encoding);
         g_key_file_set_string(keyfile, "Bookinfo", "variable_font", priv->variable_font);
         g_key_file_set_string(keyfile, "Bookinfo", "fixed_font", priv->fixed_font);
+        g_key_file_set_string(keyfile, "Bookinfo", "charset", priv->charset);
 
         gsize    length = 0;
         gchar *contents = g_key_file_to_data(keyfile, &length, NULL);
@@ -829,23 +873,25 @@ save_bookinfo(CsChmfile *self)
 }
 
 static gchar *
-check_file_ncase(CsChmfile *self, gchar *path)
+check_file_ncase(CsChmfile *self, const gchar *file)
 {
         CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
-        gchar *filename = g_build_filename(priv->bookfolder, path, NULL);
+        g_debug("CS_CHMFILE >>> check ncase file = %s", file);
+
+        gchar *filename = g_build_filename(priv->bookfolder, file, NULL);
+        gchar *new_file = NULL;
 
         if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
-                g_free(path);
-                path = NULL;
-
                 gchar *found = file_exist_ncase(filename);
                 if (found) {
-                        path = g_path_get_basename(found);
+                        gchar *basename = g_path_get_basename(found);
+                        new_file = g_strdup(basename);
+                        g_free(basename);
                         g_free(found);
                 }
         }
         g_free(filename);
-        return path;
+        return new_file;
 }
 
 /* see http://code.google.com/p/chmsee/issues/detail?id=12 */
@@ -867,6 +913,20 @@ extract_post_file_write(const gchar *fname)
         g_free(basename);
 }
 
+static void
+parse_filename(CsChmfile *self, const gchar *filename)
+{
+        CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
+        gchar *p = g_strrstr(filename, "::");
+        if (p) {
+                priv->chm = g_strndup(filename, p - filename);
+                priv->page = g_strdup(p + 2);
+        } else {
+                priv->chm = g_strdup(filename);
+        }
+        g_debug("CS_CHMFILE >>> priv->chm = %s, priv->page = %s", priv->chm, priv->page);
+}
+
 /* External functions */
 
 CsChmfile *
@@ -875,14 +935,15 @@ cs_chmfile_new(const gchar *filename, const gchar *bookshelf)
         CsChmfile        *self = g_object_new(CS_TYPE_CHMFILE, NULL);
         CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
 
+        parse_filename(self, filename);
+
         /* use chm file's MD5 as folder name */
-        gchar *md5 = MD5File(filename, NULL);
+        gchar *md5 = MD5File(priv->chm, NULL);
         if(!md5) {
                 g_warning("CS_CHMFILE >>> Oops!! cannot calculate chmfile's MD5 value.");
                 return NULL;
         }
 
-        priv->filename = g_strdup(filename);
         priv->bookfolder = g_build_filename(bookshelf, md5, NULL);
         g_debug("CS_CHMFILE >>> book folder = %s", priv->bookfolder);
 
@@ -890,8 +951,48 @@ cs_chmfile_new(const gchar *filename, const gchar *bookshelf)
         if (g_file_test(priv->bookfolder, G_FILE_TEST_IS_DIR)) {
                 load_bookinfo(self);
         } else {
-                if (!extract_chm(filename, priv->bookfolder)) {
-                        g_warning("CS_CHMFILE >>> extract_chm failed: %s", filename);
+                GtkWidget *popup_window = gtk_window_new(GTK_WINDOW_POPUP);
+                GtkWidget *progress = gtk_progress_bar_new();
+
+                gtk_container_add(GTK_CONTAINER (popup_window), progress);
+                gtk_window_set_position(GTK_WINDOW (popup_window), GTK_WIN_POS_CENTER);
+                gtk_widget_set_size_request(popup_window, 300, 40);
+                gtk_widget_show_all(popup_window);
+
+                struct extract_data data;
+
+                data.filename = priv->chm;
+                data.bookfolder = priv->bookfolder;
+                data.rval = -1;
+
+                g_thread_create((GThreadFunc) (extract_file), &data, FALSE, NULL);
+
+                /* display a faked progress bar */
+                gdouble percent = 0.0;
+                while (data.rval < 0 && percent <= 100.0) {
+                        gchar *message = g_strdup_printf(_("Processing... %.0f%% complete"), percent);
+                        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress), percent / 100.0);
+                        gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), message);
+
+                        while (gtk_events_pending())
+                                gtk_main_iteration();
+
+                        g_usleep(100000);
+                        percent += 5.0;
+                        g_free(message);
+                }
+
+                gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR (progress), 1.0);
+                gtk_progress_bar_set_text(GTK_PROGRESS_BAR (progress), _("Processing done. Loading..."));
+
+                while (gtk_events_pending())
+                        gtk_main_iteration();
+                g_usleep(200000);
+
+                gtk_widget_destroy(GTK_WIDGET (popup_window));
+
+                if (data.rval == 1) {
+                        g_warning("CS_CHMFILE >>> extract_chm failed: %s", priv->chm);
                         return NULL;
                 }
 
@@ -901,7 +1002,6 @@ cs_chmfile_new(const gchar *filename, const gchar *bookshelf)
 
         g_debug("CS_CHMFILE >>> priv->hhc = %s", priv->hhc);
         g_debug("CS_CHMFILE >>> priv->hhk = %s", priv->hhk);
-        g_debug("CS_CHMFILE >>> priv->hhk p= %p", priv->hhk);
         g_debug("CS_CHMFILE >>> priv->homepage = %s", priv->homepage);
         g_debug("CS_CHMFILE >>> priv->bookname = %s", priv->bookname);
         g_debug("CS_CHMFILE >>> priv->endcoding = %s", priv->encoding);
@@ -910,17 +1010,20 @@ cs_chmfile_new(const gchar *filename, const gchar *bookshelf)
         if (priv->hhc != NULL && g_ascii_strcasecmp(priv->hhc, "(null)") != 0) {
                 gchar *hhcfile = g_build_filename(priv->bookfolder, priv->hhc, NULL);
 
-                priv->toc_tree = parse_hhc_file(hhcfile, priv->encoding);
+                priv->toc_tree = cs_parse_file(hhcfile, priv->encoding);
+                priv->toc_list = convert_node_to_list(priv->toc_tree);
                 g_free(hhcfile);
         }
 
         /* parse hhk file */
         if (priv->hhk != NULL && priv->index_list == NULL) {
-                gchar *path = g_build_filename(priv->bookfolder, priv->hhk, NULL);
+                gchar *hhkfile = g_build_filename(priv->bookfolder, priv->hhk, NULL);
 
-                parse_hhk_file(self, path, priv->encoding);
-                g_debug("CS_CHMFILE >>> priv->index_list = %p", priv->index_list);
-                g_free(path);
+                GNode *tree = cs_parse_file(hhkfile, priv->encoding);
+                priv->index_list = convert_node_to_list(tree);
+
+                g_node_destroy(tree);
+                g_free(hhkfile);
         }
 
         /* Load bookmarks */
@@ -939,6 +1042,12 @@ cs_chmfile_get_toc_tree(CsChmfile *self)
         return CS_CHMFILE_GET_PRIVATE (self)->toc_tree;
 }
 
+GList *
+cs_chmfile_get_toc_list(CsChmfile *self)
+{
+        return CS_CHMFILE_GET_PRIVATE (self)->toc_list;
+}
+
 GList *
 cs_chmfile_get_index_list(CsChmfile *self)
 {
@@ -969,7 +1078,7 @@ cs_chmfile_update_bookmarks_list(CsChmfile *self, GList *links)
 const gchar *
 cs_chmfile_get_filename(CsChmfile *self)
 {
-        return CS_CHMFILE_GET_PRIVATE (self)->filename;
+        return CS_CHMFILE_GET_PRIVATE (self)->chm;
 }
 
 const gchar *
@@ -992,6 +1101,14 @@ cs_chmfile_get_homepage(CsChmfile *self)
         return CS_CHMFILE_GET_PRIVATE (self)->homepage;
 }
 
+const gchar *
+cs_chmfile_get_page(CsChmfile *self)
+{
+        g_return_val_if_fail(IS_CS_CHMFILE (self), NULL);
+
+        return CS_CHMFILE_GET_PRIVATE (self)->page;
+}
+
 const gchar *
 cs_chmfile_get_variable_font(CsChmfile *self)
 {
@@ -1025,3 +1142,19 @@ cs_chmfile_set_fixed_font(CsChmfile *self, const gchar *font_name)
 
         priv->fixed_font = g_strdup(font_name);
 }
+
+const gchar *
+cs_chmfile_get_charset(CsChmfile *self)
+{
+        return CS_CHMFILE_GET_PRIVATE (self)->charset;
+}
+
+void
+cs_chmfile_set_charset(CsChmfile *self, const gchar *charset)
+{
+        CsChmfilePrivate *priv = CS_CHMFILE_GET_PRIVATE (self);
+
+        g_free(priv->charset);
+
+        priv->charset = g_strdup(charset);
+}
